import json
import AlteryxPythonSDK as Sdk
import connectors_core.alteryx as alteryx
import connectors_core.alteryx.helpers as alteryx_helpers

from connectors_core.dynamics import DynamicsCrmApiClient
from connectors_core.oauth2 import OAUTH_CODE_WORKFLOW, OAUTH_CLIENT_CREDENTIALS_WORKFLOW
from connectors_core.odata.entities import EntitySchemaReader
from dynamics_crm_input.common import INT32_MAX_VALUE, DEFAULT_MAX_FIELD_LENGTH
from odata import ODataError

from typing import Union


class DynamicsCrmInputEngine(alteryx.AyxPluginBase):

    def __init__(self, n_tool_id: int, alteryx_engine: Sdk.AlteryxEngine, output_anchor_mgr: Sdk.OutputAnchorManager):
        super().__init__(n_tool_id, alteryx_engine, output_anchor_mgr)
        self.org_uri = None  # type: str
        self.tenant = None  # type: str
        self.client_id = None  # type: str
        self.client_secret = None  # type: str
        self.entity_name = None  # type: str
        self.token = None  # type: dict
        self.record_limit = 0  # 0 = No limit
        self.cached_metadata = None  # type: str
        self.max_field_length = 255
        self.custom_filter = None  # type: str
        self.using_custom_filter = None # type: bool
        self.auth_workflow_mode = OAUTH_CODE_WORKFLOW

    def pi_add_incoming_connection(self, str_type: str, str_name: str) -> Union[alteryx.IncomingConnectionBase, None]:
        self.display_error_msg(self.alteryx_engine.xmsg("This tool does not accept incoming connections"))
        return None

    def pi_init(self, str_xml: str):
        super().pi_init(str_xml)

        self.entity_name = self.get_xml_config_value(str_xml, 'inputEntityName')
        self.custom_filter = self.get_xml_config_value(str_xml, 'inputCustomFilter')
        self.using_custom_filter = self.get_xml_config_value(str_xml, 'usingCustomFilter') == 'True'
        self.org_uri = self.get_xml_config_value(str_xml, 'authOrganizationUri')
        self.tenant = self.get_xml_config_value(str_xml, 'authTenant')
        self.client_id = self.get_xml_config_value(str_xml, 'authClientId')
        self.client_secret = self.get_encrypted_xml_config_value(str_xml, 'authClientSecret')
        self.cached_metadata = self.get_xml_config_value(str_xml, 'serviceMetadata')
        self.auth_workflow_mode = self.get_xml_config_value(str_xml, 'authType')

        if self.using_custom_filter and not self.custom_filter:
            self.invalidate_config(self.alteryx_engine.xmsg('Please enter a non-empty custom filter.'))

        should_limit_records = self.get_xml_config_value(str_xml, 'shouldLimitRecords')
        if should_limit_records == 'True':
            try:
                self.record_limit = int(self.get_xml_config_value(str_xml, 'inputRecordLimit'))
                if self.record_limit < 1 or self.record_limit >= INT32_MAX_VALUE:
                    self.invalidate_config(self.alteryx_engine.xmsg(
                        "Please select a value between 1 and @1 for the record limit", str(INT32_MAX_VALUE - 1)
                    ))
            except (ValueError, TypeError) as e:
                self.record_limit = 0

        try:
            self.max_field_length = int(self.get_xml_config_value(str_xml, 'inputMaxFieldLength'))
        except (ValueError, TypeError):
            self.max_field_length = DEFAULT_MAX_FIELD_LENGTH

        if not self.entity_name:
            self.invalidate_config(self.alteryx_engine.xmsg('Please select an Entity'))

        if not self.org_uri:
            self.invalidate_config(self.alteryx_engine.xmsg('Please enter an Organization URI'))

        if not self.tenant:
            self.invalidate_config(self.alteryx_engine.xmsg('Please enter a Tenant ID'))
        if not self.client_id:
            self.invalidate_config(self.alteryx_engine.xmsg('Please enter a Client ID'))
        if not self.client_secret:
            self.invalidate_config(self.alteryx_engine.xmsg('Please enter a Client Secret'))
        if self.auth_workflow_mode not in (OAUTH_CODE_WORKFLOW, OAUTH_CLIENT_CREDENTIALS_WORKFLOW):
            self.invalidate_config(self.alteryx_engine.xmsg("Invalid Auth Mode: @1", self.auth_workflow_mode))

        token_str = self.get_encrypted_xml_config_value(str_xml, 'authToken')

        if not token_str:
            self.invalidate_config(self.alteryx_engine.xmsg('Please log in'))
        elif self.is_initialized:
            try:
                self.token = json.loads(token_str)
                if 'expires_at' not in self.token or 'resource' not in self.token or 'access_token' not in self.token:
                    self.invalidate_config(self.alteryx_engine.xmsg('Invalid auth token'))
                if self.auth_workflow_mode == OAUTH_CODE_WORKFLOW:
                    if 'refresh_token' not in self.token:
                        self.invalidate_config(self.alteryx_engine.xmsg('Invalid auth token'))
            except Exception as e:
                self.invalidate_config(self.alteryx_engine.xmsg("Failed parsing authentication: @1", str(e)))
        
        metadata_dict = {
            'Host': self.org_uri,
            'Table': self.entity_name,
            'ServiceName': 'Dynamics CRM',
            'PluginName': 'Dynamics CRM'
        }
        self.alteryx_engine.create_connect_metadata(self.n_tool_id, metadata_dict)

        self.output_anchor = self.output_anchor_mgr.get_output_anchor('Output')  # type: Sdk.OutputAnchor

    def pi_add_outgoing_connection(self, str_name: str) -> bool:

        return True

    def pi_push_all_records(self, n_record_limit: int = -1) -> bool:
        if not self.is_initialized:
            return False

        if n_record_limit >= INT32_MAX_VALUE:
            # Dynamics CRM only allows a limit value up to the max value for a 32 bit integer,
            # which effectively makes any n_record_limit value exceeding this maximum a limitless query
            n_record_limit = -1

        update_only = self.alteryx_engine.get_init_var(self.n_tool_id, 'UpdateOnly') == 'True' \
            or n_record_limit == 0

        api = DynamicsCrmApiClient(
            self.org_uri,
            self.tenant,
            self.client_id,
            self.client_secret,
            self.token,
            self.cached_metadata,
            self.auth_workflow_mode,
            alteryx_engine=self.alteryx_engine
        )

        entity_cls = api.get_entity_cls(self.entity_name)
        query = api.query(entity_cls)

        schema_reader = EntitySchemaReader(self.alteryx_engine, entity_cls, self.max_field_length)

        # Update Record info
        record_info = schema_reader.get_record_info()
        self.output_anchor.init(record_info)

        if update_only:
            return True

        record_creator = record_info.construct_record_creator()  # type: Sdk.RecordCreator
        entity_fields = schema_reader.get_entity_fields()

        if self.record_limit > 0 or n_record_limit > 0:
            query = query.limit(self.record_limit if self.record_limit > 0 else n_record_limit)

        if self.custom_filter and len(self.custom_filter.strip()) > 0:
            query = query.filter(self.custom_filter.strip())

        # enable total count to be returned with query
        query = query.inlinecount()

        record_count = 0
        try:
            for result in query:
                record_info_out = record_info.clone()  # type: Sdk.RecordInfo
                record_creator.reset()
                for f in entity_fields:
                    idx = record_info_out.get_field_num(f['name'])
                    field = record_info_out[idx]
                    field_value = getattr(result, f['name'])
                    alteryx_helpers.set_value_on_field(field, record_creator, field_value, self.alteryx_engine)

                finalized_record = record_creator.finalize_record()
                self.output_anchor.push_record(finalized_record)

                record_count += 1
                self.output_anchor.update_progress(record_count / result.__odata__.count)

            self.display_info_msg(self.alteryx_engine.xmsg(
                "@1 @2 records were read from @3", str(record_count), self.entity_name, self.org_uri
            ))
            self.output_anchor.update_progress(1.0)
            self.output_anchor.close()  # Close outgoing connections.

            return True
        except ODataError as e:
            self.log.error('OData error in pi_push_all_records: %s', e)
            self.display_error_msg(self.alteryx_engine.xmsg(
                "There was a problem querying the Dynamics CRM OData service. Code: @1", str(e.code)
            ))
            return False
        except ConnectionError as ce:
            self.log.error('Connection error in pi_push_all_records: %s', ce)
            self.display_error_msg(self.alteryx_engine.xmsg(
                "Could not connect to the Dynamics CRM OData service.  Please check your network connection"
            ))
            return False

    def pi_close(self, b_has_errors: bool):
        super().pi_close(b_has_errors)
        if self.output_anchor:
            self.output_anchor.assert_close()
