import AlteryxPythonSDK as Sdk
from typing import Type

import connectors_core.alteryx.helpers as alteryx_helpers
from connectors_core import alteryx
from connectors_core.dynamics import DynamicsCrmApiClient
from connectors_core.odata.entities import EntitySchemaReader
from dynamics_crm_output.common import OUTPUTMODE_UPSERT, OUTPUTMODE_DELETE
from odata import ODataError
from odata.entity import EntityBase
from odata.property import PropertyBase


class DynamicsCrmIncomingConnectionHandler(alteryx.IncomingConnectionBase):

    def __init__(self, parent: alteryx.AyxPluginBase):
        super().__init__(parent)
        self.record_info = None  # type: Sdk.RecordInfo
        self.odata_client = None  # type: DynamicsCrmApiClient
        self.entity_cls = None  # type: Type[EntityBase]
        self.entity_schema = None  # type: EntitySchemaReader
        self.entity_name = None
        self.invalid_field_names = None
        self.output_mode = None

    def ii_init(self, record_info_in: Sdk.RecordInfo) -> bool:
        if not self.parent.is_initialized:
            return False
        self.record_info = record_info_in
        self.output_mode = self.parent.output_mode
        if self.output_mode not in (OUTPUTMODE_UPSERT, OUTPUTMODE_DELETE):
            self.parent.display_error_msg(self.parent.alteryx_engine.xmsg("Invalid output mode: @1", self.output_mode))
            return False
        self.entity_name = self.parent.entity_name
        if not self.entity_name:
            self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                "Missing entity name. Incoming connection handler could not be initialized."
            ))
            return False
        self.odata_client = DynamicsCrmApiClient(
            self.parent.org_uri,
            self.parent.tenant,
            self.parent.client_id,
            self.parent.client_secret,
            self.parent.token,
            self.parent.cached_metadata,
            self.parent.auth_workflow_mode,
            self.parent.alteryx_engine
        )
        self.entity_cls = self.odata_client.get_entity_cls(self.entity_name)
        if not self.entity_cls:
            self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                "Entity '@1' does not exist in the service metadata", self.entity_name
            ))
            return False
        self.entity_schema = EntitySchemaReader(self.parent.alteryx_engine, self.entity_cls)
        self.invalid_field_names = []
        return True

    def ii_push_record(self, in_record: Sdk.RecordRef) -> bool:
        entity = self.entity_cls()
        if self.output_mode == OUTPUTMODE_UPSERT:
            for i in range(0, self.record_info.num_fields):
                field = self.record_info[i]  # type: Sdk.Field
                try:
                    if field.name not in self.invalid_field_names:
                        err_handler = lambda field, ref, err: None
                        value = alteryx_helpers.get_value_for_field(field, in_record, data_conversion_error_handler=err_handler)
                        if value is not None:
                            if not field.name.startswith('_'):
                                setattr(entity, field.name, value)
                            else:
                                setattr(entity, '{}@odata.bind'.format(field.name), value)
                except AttributeError as e:
                    self.log.warning("Error setting attribute on entity: %s", e)
                    self.parent.display_warning_msg(self.parent.alteryx_engine.xmsg(
                        "The field @1 does not exist on the entity @2", field.name, self.entity_name
                    ))
                    self.invalid_field_names.append(field.name)
            try:
                self.odata_client.save(entity)
            except ODataError as e:
                self.log.error("Failed to save entity {}".format(entity), e)
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg("Failed to save entity: @1", str(e.code)))
            except ConnectionError as ce:
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                    "Unable to connect to the Dynamics CRM service, please check your network connection."
                ))
                self.log.error(ce)
                raise
            except Exception as ex:
                self.log.error("Exception encountered saving entity {}".format(entity), ex)
                raise
        elif self.output_mode == OUTPUTMODE_DELETE:
            primary_keys = self.entity_schema.get_primary_key_fields()
            if not primary_keys:
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                    "Entity @1 is missing a primary_key field.  Delete operation is not possible", self.entity_name
                ))
                return False
            elif len(primary_keys) == 1:
                try:
                    self._set_prop_value_from_record(entity, next(iter(primary_keys)), in_record)
                except AttributeError as e:
                    self.log.error("Failed setting primary key value on entity: %s", e)
                    self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                        "Failed to set primary key value on entity: @1", str(e)
                    ))
                    return False
            else:  # iterable of primary keys
                try:
                    for pk in primary_keys:
                        self._set_prop_value_from_record(entity, pk, in_record)
                except AttributeError as e:
                    self.log.error("Failed setting primary key value on entity: %s", e)
                    self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                        "One or more primary key fields could not be set from RecordInfo:\n@1",
                        (",".join([pk.name for pk in primary_keys]))
                    ))
                    return False
            try:
                self.odata_client.delete(entity)
            except ODataError as e:
                self.log.error("Failed to delete entity {}".format(entity), e)
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                    "Failed to delete entity @1: @2", entity, e.message
                ))
            except ConnectionError as ce:
                self.log.error(ce)
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                    "Unable to connect to the Dynamics CRM service, please check your network connection."
                ))
                raise
            except Exception as ex:
                self.log.error("Exception thrown while deleting entity {}".format(entity), ex)
                raise
        return True

    def _set_prop_value_from_record(self, entity: EntityBase, prop: PropertyBase, in_record: Sdk.RecordRef):
        field = self.record_info.get_field_by_name(prop.name)
        if not field:
            raise AttributeError("{} does not exist in record_info".format(prop.name))
        setattr(entity, prop.name, alteryx_helpers.get_value_for_field(field, in_record))

    def ii_update_progress(self, d_percent: float):
        try:
            self.parent.alteryx_engine.output_tool_progress(self.parent.n_tool_id, d_percent)
        except AttributeError:
            pass

    def ii_close(self):
        self.is_initialized = False
