import json
import AlteryxPythonSDK as Sdk
import connectors_core.alteryx as alteryx

from typing import Union

from dynamics_crm_output.common import OAUTH_CODE_WORKFLOW_MODE, OAUTH_CLIENT_CREDENTIALS_WORKFLOW_MODE, \
    OUTPUTMODE_UPSERT, OUTPUTMODE_DELETE
from .connections import DynamicsCrmIncomingConnectionHandler


class DynamicsCrmOutputEngine(alteryx.AyxPluginBase):

    def __init__(self, n_tool_id: int, alteryx_engine: Sdk.AlteryxEngine, output_anchor_mgr: Sdk.OutputAnchorManager):
        super().__init__(n_tool_id, alteryx_engine, output_anchor_mgr)
        self.org_uri = None  # type: str
        self.tenant = None  # type: str
        self.client_id = None  # type: str
        self.client_secret = None  # type: str
        self.entity_name = None  # type: str
        self.token = None  # type: dict
        self.cached_metadata = None  # type: str
        self.output_mode = None  # type: str
        self.auth_workflow_mode = OAUTH_CODE_WORKFLOW_MODE
        self.incoming_connection_added = False

    def pi_add_incoming_connection(self, str_type: str, str_name: str) -> Union[alteryx.IncomingConnectionBase, None]:
        if self.incoming_connection_added:
            self.display_error_msg(self.alteryx_engine.xmsg(
                "Only one incoming connection can be accepted by this tool"
            ))
            return None
        connection_handler = DynamicsCrmIncomingConnectionHandler(self)
        self.incoming_connection_added = True
        return connection_handler

    def pi_init(self, str_xml: str):
        super().pi_init(str_xml)

        self.entity_name = self.get_xml_config_value(str_xml, 'outputEntityName')
        self.org_uri = self.get_xml_config_value(str_xml, 'authOrganizationUri')
        self.tenant = self.get_xml_config_value(str_xml, 'authTenant')
        self.client_id = self.get_xml_config_value(str_xml, 'authClientId')
        self.client_secret = self.get_encrypted_xml_config_value(str_xml, 'authClientSecret')
        self.cached_metadata = self.get_xml_config_value(str_xml, 'serviceMetadata')
        self.auth_workflow_mode = self.get_xml_config_value(str_xml, 'authType')
        self.output_mode = self.get_xml_config_value(str_xml, 'outputMode')

        if not self.entity_name:
            self.invalidate_config(self.alteryx_engine.xmsg('Please select an Entity'))

        if not self.org_uri:
            self.invalidate_config(self.alteryx_engine.xmsg('Please enter an Organization URI'))

        if not self.tenant:
            self.invalidate_config(self.alteryx_engine.xmsg('Please enter a Tenant ID'))
        if not self.client_id:
            self.invalidate_config(self.alteryx_engine.xmsg('Please enter a Client ID'))
        if not self.client_secret:
            self.invalidate_config(self.alteryx_engine.xmsg('Please enter a Client Secret'))

        if self.auth_workflow_mode not in (OAUTH_CODE_WORKFLOW_MODE, OAUTH_CLIENT_CREDENTIALS_WORKFLOW_MODE):
            self.invalidate_config(self.alteryx_engine.xmsg("Invalid Auth Mode: @1", self.auth_workflow_mode))

        if not self.output_mode:
            self.invalidate_config(self.alteryx_engine.xmsg("Please select an output operation"))
        elif self.output_mode not in (OUTPUTMODE_UPSERT, OUTPUTMODE_DELETE):
            self.invalidate_config(self.alteryx_engine.xmsg("Invalid output operation: @1", self.output_mode))

        token_str = self.get_encrypted_xml_config_value(str_xml, 'authToken')

        if not token_str:
            self.invalidate_config(self.alteryx_engine.xmsg('Please log in'))
        elif self.is_initialized:
            try:
                self.token = json.loads(token_str)
                if 'expires_at' not in self.token or 'resource' not in self.token or 'access_token' not in self.token:
                    self.invalidate_config(self.alteryx_engine.xmsg("Invalid auth token format: @1", token_str))
                if self.auth_workflow_mode == OAUTH_CODE_WORKFLOW_MODE:
                    if 'refresh_token' not in self.token:
                        self.invalidate_config(self.alteryx_engine.xmsg(
                            "Invalid auth token format for user auth: @1", token_str
                        ))
            except Exception as e:
                self.invalidate_config(self.alteryx_engine.xmsg("Failed parsing authentication: @1", str(e)))

        
        metadata_dict = {
            'Host': self.org_uri,
            'Table': self.entity_name,
            'ServiceName': 'Dynamics CRM',
            'PluginName': 'Dynamics CRM'
        }
        self.alteryx_engine.create_connect_metadata(self.n_tool_id, metadata_dict)

        self.output_anchor = self.output_anchor_mgr.get_output_anchor('Output')  # type: Sdk.OutputAnchor

    def pi_add_outgoing_connection(self, str_name: str) -> bool:
        self.display_error_msg(self.alteryx_engine.xmsg('This tool does not allow outgoing connections'))

        return False

    def pi_push_all_records(self, n_record_limit: int) -> bool:
        return False

    def pi_close(self, b_has_errors: bool):
        super().pi_close(b_has_errors)
        if self.output_anchor:
            self.output_anchor.assert_close()
